#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "global.h"
#include "functions.h"
#include "util.h"
#include "ranlib.h"
#include "our_getopt.h"
#include "plist.h"
#include "blackScholes.h"
#include "plist.h"


double optionPayoff(double *price){
  int maxPrice;

  maxPrice = dArrayMax(Dimensions,price);
  return(maxPrice-Strike_price);
}

double innerControlledPathEstimate(void){
  int time, underlier, toPath, count,max1,max2,i;
  double **path, **antiPath, antiGrowth, *payoff, *antiPayoff, *valIfHeld,
    *antiValIfHeld, val, antiVal,  delta, euroVal,
    maxval1, maxval2, var, error;
  POINT scratch_point, *mesh_address;


  allocate2dDouble(&path, Timeslices, Dimensions);
  allocate2dDouble(&antiPath,  Timeslices, Dimensions);
  allocate1dDouble(&payoff, Timeslices);
  allocate1dDouble(&antiPayoff, Timeslices);
  allocate1dDouble(&valIfHeld, Timeslices);
  allocate1dDouble(&antiValIfHeld, Timeslices);
  
  
  val = 0;
  antiVal = 0;
  
  for(count=1; count<=Path_estimate_paths; count++){
    pathPair(path,antiPath);

    for(time=0; time < Timeslices; time++){
      payoff[time] = exp(-time*Interest_rate)*
        optionPayoff(path[time]);
      antiPayoff[time] = exp(-time*Interest_rate)*
        optionPayoff(antiPath[time]);
    }

    valIfHeld[0] = dMax(0,payoff[0]);
    antiValIfHeld[0] = dMax(0,antiPayoff[0]);

    valIfHeld[Timeslices-1] = 0;
    antiValIfHeld[Timeslices-1] = 0;

    for(time=0; time < Timeslices; time++){
      /* We don't need to compute val if held on the last timeslice */
      if(time == Timeslices-1){
        if(payoff[time] > valIfHeld[time]){
          val += payoff[time];
          break;
        }
      }else{
        /* Skip point if payoff not > 0 */
        if(payoff[time] > 0 ){
          max1 = dWhichMax(Dimensions, path[time]); 
          euroVal = exp(-time*Interest_rate)*
            euroCall(path[time][max1],
                     Strike_price,
                     Interest_rate,
                     sqrt(Covariance[max1][max1]),
                     Timeslices-1 - time,
                     Dividend_rate[max1]);
          /* Skip point if payoff not > 1d european */
          if(payoff[time] > euroVal){
            maxval1 = path[time][max1];
            maxval2 = 0;
            for(i = 0; i<Dimensions; i++){
              if(i != max1 && path[time][i] > maxval2){
                max2 = i;
                maxval2 = path[time][i];
              }
            } 
            euroVal = exp(-time*Interest_rate)* 
              maxOf2Risk(maxval1, maxval2, Strike_price,
                         Timeslices-1 - time,
                         sqrt(Covariance[max1][max1]), 
                         sqrt(Covariance[max2][max2]), 
                         Interest_rate, 
                         Dividend_rate[max1], 
                         Dividend_rate[max2],
                         Covariance[max1][max2]);
            /* Skip point if payoff not > 2d european */
            if(payoff[time] > euroVal){
	      point_initialize(&scratch_point);
	      for(underlier=0; underlier<Dimensions;underlier++){
		scratch_point.price[underlier] = path[time][underlier];
	      }
	      scratch_point.timeslice = time;
	      find_important_clusters(&scratch_point);
              valIfHeld[time] = mesh_estimate_value(&scratch_point,&error);
	      
	      if(fabs(error) > fabs(scratch_point.delta_r)*Error_target && error > Error_floor){
		mesh_address = add_point_to_mesh(&scratch_point,NULL,time);
		value(NULL,mesh_address);
	      }
	      point_finalize(&scratch_point);
	      
              if(payoff[time] > valIfHeld[time]){
                val += payoff[time];
                break;
              }
            }
          }
        }
      }
    }

    for(time=0; time < Timeslices; time++){
      if(time == Timeslices-1){
        if(antiPayoff[time] > antiValIfHeld[time]){
          antiVal += antiPayoff[time];
          break;
        }
      }else{
        if(antiPayoff[time] > 0 ){
          max1 = dWhichMax(Dimensions, antiPath[time]);
          euroVal = exp(-time*Interest_rate)*
            euroCall(antiPath[time][max1],
                     Strike_price,
                     Interest_rate,
                     sqrt(Covariance[max1][max1]),
                     Timeslices-1 - time,
                     Dividend_rate[max1]);
          if(antiPayoff[time] > euroVal){
            maxval1 = antiPath[time][max1];
            maxval2 = 0;
            for(i = 0; i<Dimensions; i++){
              if(i != max1 && antiPath[time][i] > maxval2){
                max2 = i;
                maxval2 = antiPath[time][i];
              }
            } 
            euroVal = exp(-time*Interest_rate)* 
              maxOf2Risk(maxval1, maxval2, Strike_price,
                         Timeslices-1 - time,
                         sqrt(Covariance[max1][max1]), 
                         sqrt(Covariance[max2][max2]), 
                         Interest_rate, 
                         Dividend_rate[max1], 
                         Dividend_rate[max2],
                         Covariance[max1][max2]);
            if(antiPayoff[time] > euroVal){
	      point_initialize(&scratch_point);
	      for(underlier=0; underlier<Dimensions;underlier++){
		scratch_point.price[underlier] = antiPath[time][underlier];
	      }
	      scratch_point.timeslice = time;
	      find_important_clusters(&scratch_point);
              antiValIfHeld[time] = mesh_estimate_value(&scratch_point,&error);
	      if(fabs(error) > fabs(scratch_point.delta_r)*Error_target && error > Error_floor){
		mesh_address = add_point_to_mesh(&scratch_point,NULL,time);
		value(NULL,mesh_address);
	      }

	      point_finalize(&scratch_point);

              if(antiPayoff[time] > antiValIfHeld[time]){
                antiVal += antiPayoff[time];
                break;
              }
            }
          }
        }
      }
    }

   }


  val /= Path_estimate_paths;
  antiVal /= Path_estimate_paths;
  
  deallocate2dDouble(&path,  Timeslices,
                     Dimensions); 
  deallocate2dDouble(&antiPath,  Timeslices,
                     Dimensions); 
  deallocate1dDouble(&payoff, Timeslices); 
  deallocate1dDouble(&antiPayoff, Timeslices); 
  deallocate1dDouble(&valIfHeld, Timeslices); 
  deallocate1dDouble(&antiValIfHeld, Timeslices); 

  return(0.5*(val+antiVal));  
}
